unit PolynomUtil;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, typinfo, Dialogs, Math;

type
 generic Polynom<Precision> = class
  type

   Complex = object
    strict private
     function ReadAbs2 : Precision; inline;
    public
     Re : Precision;
     Im : Precision;
     property Abs2 : Precision read ReadAbs2;
     function sub(c : Complex) : Complex; inline;
     function add(c : Complex) : Complex; inline;
   end;

   IFunction = class
     function TrapezIntegral(min, max : Precision; n : Cardinal) : Precision;
     function Calc(x : Precision) : Precision; virtual;
     function Newton(y : Precision; df : IFunction; var x : Precision; min : Precision = 0; max : Precision = 0; gradient : Precision = 0) : Boolean;
     function CalcInverse(y, a, b : Precision; df : IFunction) : Precision;
     function Derivate : IFunction; virtual; abstract;
     function Integrate(c : Precision) : IFunction; virtual; abstract;
     function ToString : String; virtual; abstract;
     procedure FromString(S : String); virtual; abstract;
   end;
   PFunction = ^IFunction;

   TJBlur = class(IFunction)
    strict private
     t : Precision;
     function intNC4(min, max : Precision; m : Cardinal) : Precision; inline;
     function ConvolutionVal(x : Precision) : Precision; inline;
    public
     radius : Precision;
     integrateQuality : cardinal;
     source : IFunction;
     constructor Create(_source : IFunction; blurRadius : Precision = 1);
     function Calc(x : Precision) : Precision; override;
   end;

   TFourier = class(IFunction)
    private
     fkt : IFunction;
     constructor Create(_fkt : IFunction);
     destructor destroy; override;
    public
     function Calc(x : Precision) : Precision; override;
   end;

   TPolynom = class(IFunction)
    private
     _co         : Array of Precision;
     _ze         : Array of Precision;
     _grad       : Cardinal;
     procedure calcCoefficients;
     procedure calcZeros;
     function ReadCo(i : Cardinal) : Precision; inline;
     procedure WriteCo(i : Cardinal; v : Precision);
     function FourierIntegral (w : Precision; t : Precision) : Complex; //specialize Complex<Precision>;
    public
     constructor Create;
     function ToString : String; override;
     procedure FromString(s : String); override;
     property co[i : Cardinal] : Precision read ReadCo write WriteCo; default;
     property Grad : Cardinal read _grad;
     function Calc(x : Precision) : Precision; override;
     procedure Add(P : TPolynom);
     function Derivate : TPolynom; override;
     function Integrate(c : Precision) : TPolynom; override;
     destructor Destroy; override;
   end;

   TPolySpline = class(IFunction)
    type TPolySplinePart = record
     Fkt : TPolynom;
     Min : Precision;
     Max : Precision;
    end;
    private
     _Splines   : Array of TPolySplinePart;
     _LastIndex : Cardinal;
     function ReadSpline(i : Cardinal) : TPolySplinePart; inline;
     function ReadCount : Cardinal; inline;
    public
     constructor Create;
     function ToString : String; override;
     property Splines[i : Cardinal] : TPolySplinePart read ReadSpline; default;
     property Count : Cardinal read ReadCount;
     function Calc(x : Precision) : Precision; override;
     function Derivate : TPolySpline; override;
     function Integrate(c : Precision) : TPolySpline; override;
     destructor Destroy; override;
   end;

   TInterpolateInfo = object
    type
     TInterpolatePoint = object
      strict private
       _x : Precision;
       _y : Precision;
       _derivation : cardinal;
      public
       property x : Precision read _x write _x;
       property y : Precision read _y write _y;
       property derivation : Cardinal read _derivation write _derivation;
     end;
    strict private
     arr : Array of TInterpolatePoint;
     sorted : Boolean;
     function ReadPoint(i : Cardinal) : TInterpolatePoint; inline;
     procedure WritePoint(i : Cardinal; p : TInterpolatePoint); inline;
     function ReadCount : Cardinal; inline;
     procedure WriteCount(c : Cardinal); inline;
    public
     constructor Init;
     constructor FromString(str : String);
     property count : Cardinal read ReadCount write WriteCount;
     procedure sort(_lo : integer = 0; _hi : integer = -1);
     property points[i : Cardinal] : TInterpolatePoint read ReadPoint write WritePoint; default;
   end;
   PInterpolateInfo = ^TInterpolateInfo;

   TVector = Array of Precision;

   TMatrix = object
    strict private
     Mem  : Array of Precision;
     dim1 : Cardinal;
     dim2 : Cardinal;
     cv   : TVector;
     function _get(i, j : Cardinal) : Precision; inline;
     procedure _set(i, j : Cardinal; v : Precision); inline;
    public
     constructor Init(i, j : Cardinal; Identity : Boolean);
     property Colums : Cardinal read dim2;
     property Rows : Cardinal read dim1;
     procedure SwitchRows(i, j : Cardinal);
     property Objects[i, j : Cardinal] : Precision read _get write _set; default;
     function ToString : String;
     destructor Destroy;
   end;

  private
   class var _LastError : String;
   class function FloatToStr(f : Precision) : String; inline;
   class function Trunc(f : Precision) : Integer; inline;
   class function VecToStr(v : TVector) : String;
   class function fak(v, n : Cardinal) : Cardinal;
   class function StrToIntErr(s : String; i : integer) : integer;
   class function StrToFloatErr(s : String; f : float) : float;
  public
   class function HermiteInterpolation(const IP : TInterpolateInfo) : TPolynom;
   class function NewtonInterpolation(const IP : TInterpolateInfo) : TPolynom;
   class function AdjustedSplinesOnFirstDerivation(var IP : TInterpolateInfo; AdjustSplines, AdjustPoints : Precision) : TPolySpline;
   class function Gauss(var M : TMatrix; var v : TVector; var b : TVector) : Boolean;
   class function KubicSplines(var IP : TInterpolateInfo) : TPolySpline;
   class function KubicSplines2(var IP : TInterpolateInfo; hermiteRB : Boolean; fa, fb : Precision) : TPolySpline;
   class function LinearSplines(var IP : TInterpolateInfo) : TPolySpline;
   class function IsNaN(v : Precision) : Boolean; inline;
   class function NaN : Precision; inline;
   class function LastError : String;
 end;


implementation



class function Polynom.LastError : String;
begin
 Result := _LastError;
end;

class function Polynom.VecToStr(v : TVector) : String;
var i : integer;
begin
 Result := '(';
 for i := 0 to Length(v)-1 do begin
  Result := Result + FloatToStr(v[i]);
  if (i <> Length(v)-1) then Result := Result + ' | ';
 end;
 Result := Result + ')';
end;



constructor Polynom.TInterpolateInfo.Init;
begin
 arr := nil;
 sorted := false;
end;

constructor Polynom.TInterpolateInfo.FromString(str : String);
var
 S   : AnsiString;
 i,j : integer;
 b   : Byte;
begin
 sorted := false;
 SetLength(arr,0);
 b := 0;
 for i := 1 to Length(str) do if (b = 0) then begin
  if (str[i] = '[') then begin
   b := 1;
   s := '';
  end;
 end else if (b = 1) then begin
  if (str[i] = '|') or (str[i] = ';') then begin
   j := StrToIntErr(s,-1);
   if (j < 0) or (j > 65535) then exit;
   SetLength(arr,Length(arr) + 1);
   arr[Length(arr) - 1].X := j;
   arr[Length(arr) - 1].Y := 0;
   b := 2;
   s := '';
  end else s := s + str[i];
 end else if (b = 2) then begin
  if (str[i] = ']') then begin
   arr[Length(arr) - 1].Y := StrToFloatErr(s,0);
   b := 0;
   s := '';
  end else s := s + str[i];
 end else exit;
end;

function Polynom.TInterpolateInfo.ReadPoint(i : Cardinal) : TInterpolatePoint; inline;
begin
 Result := arr[i];
end;

procedure Polynom.TInterpolateInfo.WritePoint(i : Cardinal; p : TInterpolatePoint); inline;
begin
 arr[i] := p;
 sorted := false;
end;

function Polynom.TInterpolateInfo.ReadCount : Cardinal; inline;
begin
 Result := Length(arr);
end;

procedure Polynom.TInterpolateInfo.WriteCount(c : Cardinal); inline;
begin
 SetLength(arr, c);
 sorted := false;
end;

procedure Polynom.TInterpolateInfo.sort(_lo : integer = 0; _hi : integer = -1);
var
 lo,hi : integer;
 m,s : TInterpolatePoint;
 b : boolean;
begin
 if (sorted) then exit;
 if (_hi = -1) then begin _hi := Length(arr)-1; b := true; end else b := false;
 if (Length(arr) = 0) then exit;
 m := arr[(_lo + _hi) div 2];
 lo := _lo;
 hi := _hi;
 repeat
  while (arr[lo].x < m.x) do lo := lo + 1;
  while (arr[hi].x > m.x) do hi := hi - 1;
  if (lo <= hi) then begin
   s := arr[lo];
   arr[lo] := arr[hi];
   arr[hi] := s;
   lo := lo + 1;
   hi := hi - 1;
  end;
 until (lo > hi);
 if (hi > _lo) then sort(_lo, hi);
 if (lo < _hi) then sort(lo, _hi);
 if (b) then sorted := true;
end;




function Polynom.Complex.ReadAbs2 : Precision; inline;
begin
 Result := Re*Re + Im*Im;
end;

function Polynom.Complex.sub(c : Complex) : Complex; inline;
begin
 Result.Re := self.Re-c.Re;
 Result.Im := self.Im-c.Im;
end;

function Polynom.Complex.add(c : Complex) : Complex; inline;
begin
 Result.Re := self.Re+c.Re;
 Result.Im := self.Im+c.Im;
end;


function Polynom.IFunction.Calc(x : Precision) : Precision; // to be overriden
begin
 Result := NaN;
end;

function Polynom.IFunction.TrapezIntegral(min, max : Precision; n : Cardinal) : Precision;
var
 h : Precision;
 i : integer;
 A : Precision;
begin
 Result := 0;
 h := (max - min) / n;
 for i := 1 to n do begin
  A := (Calc(min + (i-1) * h) + Calc(min + i*h)) * 0.5 * h;
  if (not IsNaN(A)) then Result := Result + A;
 end;
end;

function Polynom.IFunction.Newton(y : Precision; df : IFunction; var x : Precision; min : Precision = 0; max : Precision = 0; gradient : Precision = 0) : Boolean;
var
 s, d, f, f_, f0 : Precision;
 n : Byte;
begin
 Result := false;
 f := self.Calc(x) - y;
 n := 0;
 f0 := f*f;
 f_ := df.Calc(x);
 if (f = 0) and (f_ * Gradient <= 0) and (Gradient <> 0) then exit;
 f_ := 0;
 while (n < SizeOf(Precision) div 16 + 5) and (f <> 0) do begin
  s := df.Calc(x);
  if (s = 0) or (s * gradient < 0) then if (f_ = 0) and (min <> max) then begin
   if (x > (max + min)/2) then x := x - (max-min)/2
   else x := x + (max-min)/2;
   f_ := 1;
   continue;
  end else exit;
  s := - f / s;
  d := 2;   // d ist Amijo-Sigma, wird in erstem Schleifendurchlauf halbiert!
  repeat begin
   d := d / 2;
   if (d*s = 0) then exit;
   if ((x + d*s < min) or (x + d*s > max)) and (max <> min) then begin f_ := 2*f; continue; end;
   if (gradient <> 0) and (df.Calc(x + d*s) * gradient < 0) then begin f_ := 2*f; continue; end;
   f_ := self.Calc(x + d*s) - y;
  end until (f_*f_ <= f*f*(1-d*0.001));
  if (f = f_) then break;
  f := f_;
  x := x + d*s;
  if (n <> 0) then n := n + 1
  else if (f*f < f0/10) then n := 1;
 end;
 Result := f*f < 1E-20;
end;

function Polynom.IFunction.CalcInverse(y, a, b : Precision; df : IFunction) : Precision;
var c : Boolean;
begin
 // Ungelöstes Problem: Globalisiertes Newtonverfahren kann aus [a, b] herausspringen, auch wenn es eine Nullstelle in [a, b] gibt!
 if (df = nil) then begin c := true; df := self.Derivate; end else c := false;
 Result := (a + b) / 2;
 try
  self.Newton(y, df, Result);
  if (Result < a) or (Result > b) then Result := NaN;
 except
  Result := NaN;
 end;
 if (c) then df.free;
end;

constructor Polynom.TMatrix.Init(i, j : Cardinal; Identity : Boolean);
begin
 cv := nil;
 SetLength(Mem, i * j);
 cv := nil;
 dim2 := j;
 dim1 := i;
 while (i > 0) do begin
  i := i - 1;
  while (j > 0) do begin
   j := j - 1;
   if (Identity) and (i = j) then Mem[i * dim2 + j] := 1
   else Mem[i * dim2 + j] := 0;
  end;
 end;
end;

procedure Polynom.TMatrix.SwitchRows(i, j : Cardinal);
begin
 if (i = j) then exit;
 if (Length(cv) < Colums) then SetLength(cv, Colums);
 Move(Mem[i*Colums], cv[0], Colums*SizeOf(Precision));
 Move(Mem[j*Colums], Mem[i*Colums], Colums*SizeOf(Precision));
 Move(cv[0], Mem[j*Colums], Colums*SizeOf(Precision));
end;

function Polynom.TMatrix._get(i, j : Cardinal) : Precision; inline;
begin
 Result := Mem[i * Colums + j];
end;

procedure Polynom.TMatrix._set(i, j : Cardinal; v : Precision) inline;
begin
 Mem[i * Colums + j] := v;
end;

function Polynom.TMatrix.ToString : String;
var i, j : integer;
begin
 Result := '';
 for i := 0 to Rows-1 do begin
  for j := 0 to Colums-1 do Result := Result + FloatToStr(Mem[i * Colums + j]) + '     ';
  Result := Result + Char(13)  + Char(10);
 end;
end;

destructor Polynom.TMatrix.Destroy;
begin
 SetLength(cv, 0);
 SetLength(Mem, 0);
 dim1 := 0;
 dim2 := 0;
end;



constructor Polynom.TJBlur.Create(_source : IFunction; blurRadius : Precision = 1);
begin
 inherited Create;
 source := _source;
 radius := blurRadius;
 integrateQuality := 100;
end;

function Polynom.TJBlur.ConvolutionVal(x : Precision) : Precision; inline;
begin
 if (((t-x)/radius)**2 >= 1) then Result := 0
 else Result := source.Calc(x) * exp(- 1/(1 - ((t-x)/radius)**2))/(radius*0.44399381616807943782304892117055325);
end;

function Polynom.TJBlur.intNC4 (min, max : Precision; m : Cardinal) : Precision; inline;
var
 s, h : Precision;
 j : integer;
begin
 s := 0;
 h := (max-min)/(m*4);
 for j := 0 to m-1 do begin
  s := s + 14/45 * ConvolutionVal(min+(j*4+0)*h)
         + 64/45 * ConvolutionVal(min+(j*4+1)*h)
         + 24/45 * ConvolutionVal(min+(j*4+2)*h)
         + 64/45 * ConvolutionVal(min+(j*4+3)*h)
         + 14/45 * ConvolutionVal(min+(j*4+4)*h);
 end;
 result := h*s;
end;

function Polynom.TJBlur.Calc(x : Precision) : Precision;
begin
 t := x;
 Result := intNC4(x-radius, x+radius, integrateQuality);
end;


constructor Polynom.TPolynom.Create;
begin
 inherited;
 SetLength(_co, 1);
 _co[0] := 0;
 SetLength(_ze, 0);
 _grad := 0;
end;

function Polynom.TPolynom.FourierIntegral (w : Precision; t : Precision) : Complex;
begin
 if (w = 0) then begin
  Result.Re := 0;
  Result.Im := 0;
 end else begin
  Result.Re := (3*self.co[3]*((t*t)*(w*w) - 2) + 2*self.co[2]*t*(w*w) + self.co[1]*(w*w))*COS(t*w)/(w*w*w*w) + (self.co[3]*t*((t*t)*(w*w) - 6) + self.co[2]*((t*t)*(w*w) - 2) + self.co[1]*t*(w*w) + self.co[0]*(w*w))*SIN(t*w)/(w*w*w);
  Result.Im := ((self.co[3]*t*((t*t)*(w*w) - 6) + self.co[2]*((t*t)*(w*w) - 2) + self.co[1]*t*(w*w) + self.co[0]*(w*w))*COS(t*w)/(w*w*w) - (3*self.co[3]*((t*t)*(w*w) - 2) + 2*self.co[2]*t*(w*w) + self.co[1]*(w*w))*SIN(t*w)/(w*w*w*w));
 end;
 // hartley Result := (self.co[3]*(t*t*t*w*w*w + 3*t*t*w*w - 6*t*w - 6) + w*(self.co[2]*(t*t*w*w + 2*t*w - 2) + w*(self.co[1]*(t*w + 1) + self.co[0]*w)))*SIN(t*w)/(w*w*w*w) - (self.co[3]*(t*t*t*w*w*w - 3*t*t*w*w - 6*t*w + 6) + w*(self.co[2]*(t*t*w*w - 2*t*w - 2) + w*(self.co[1]*(t*w - 1) + self.co[0]*w)))*COS(t*w)/(w*w*w*w);
end;

procedure Polynom.TPolynom.WriteCo(i : Cardinal; v : Precision);
begin
 if (i > _grad) then begin
  SetLength(_co, i+1);
  repeat begin _grad := _grad + 1; _co[_grad] := 0; end until (_grad >= i);
 end;
 _co[i] := v;
 if (i = _grad) and (v = 0) then while (_grad > 0) and (_co[_grad] = 0) do begin
  SetLength(_co, _grad);
  _grad := _grad - 1;
 end;
end;

function Polynom.TPolynom.ReadCo(i : Cardinal) : Precision; inline;
begin
 if (i > _grad) then Result := 0
 else Result := _co[i];
end;

procedure Polynom.TPolynom.calcCoefficients;
begin

end;

procedure Polynom.TPolynom.calcZeros;
begin

end;

procedure Polynom.TPolynom.FromString(s : String);
begin
 ShowMessage('FromString ist noch nicht fertig!');
end;

function Polynom.TPolynom.ToString : String;
var i : integer;
begin
 Result := '';
 //if (not _coValid) then calcCoefficients;
 i := Length(_co)-1;
 for i := i downto 0 do begin
  if (_co[i] < 0) then Result := Result + ' - ' + FloatToStr(-_co[i]) + ' X^' + IntToStr(i)
  else Result := Result + ' + ' + FloatToStr(_co[i]) + ' X^' + IntToStr(i);
 end;
end;

function Polynom.TPolynom.Calc(x : Precision) : Precision;
var
 Ex : Precision;
 i  : integer;
begin
 Ex := 1;
 Result := 0;
 for i := 0 to _grad do begin
  Result := Result + _co[i] * Ex;
  Ex := Ex * x;
 end;
end;

procedure Polynom.TPolynom.Add(P : TPolynom);
var i : integer;
begin
 if (P._grad > _grad) then begin
  SetLength(_co, P._grad+1);
  repeat begin _grad := _grad + 1; _co[_grad] := 0; end until (_grad >= P._grad);
 end;
 for i := P._grad downto 0 do _co[i] := _co[i] + P._co[i];
end;

function Polynom.TPolynom.Derivate : TPolynom;
var i : integer;
begin
 Result := TPolynom.Create;
 for i := _grad downto 1 do Result[i-1] := _co[i]*i;
end;

function Polynom.TPolynom.Integrate(c : Precision) : TPolynom;
var i : integer;
begin
 Result := TPolynom.Create;
 for i := _grad downto 0 do Result[i+1] := _co[i]/(i+1);
 Result[0] := c;
end;

destructor Polynom.TPolynom.Destroy;
begin
 SetLength(_co, 0);
 SetLength(_ze, 0);
 inherited;
end;




constructor Polynom.TFourier.Create(_fkt : IFunction);
begin
 inherited Create;
 fkt := _fkt;
end;

function Polynom.TFourier.Calc(x:Precision) : Precision;
var
 F : Complex;
 i, j : integer;
begin
 F.Re := 0;
 F.Im := 0;
 if (fkt is TPolySPline) then for i := 0 to TPolySpline(fkt).count-1 do begin
  F := F.add(TPolySpline(fkt)[i].Fkt.FourierIntegral(x, TPolySpline(fkt)[i].Max).sub(TPolySpline(fkt)[i].Fkt.FourierIntegral(x, TPolySpline(fkt)[i].Min)));
 end;
 Result := F.Abs2;
end;

destructor Polynom.TFourier.Destroy;
begin
 inherited;
end;



constructor Polynom.TPolySpline.Create;
begin
 inherited;
 _LastIndex := 0;
end;

function Polynom.TPolySpline.Calc(x : Precision) : Precision;
var
  i    : Cardinal;
  b    : Boolean;
  l, u : Cardinal;
begin
 Result := NaN;
 if (Length(_Splines) = 0) then exit;
 i := _LastIndex;
 l := 0;
 u := Length(_Splines)-1;
 if (i >= Length(_Splines)) then i := (Length(_Splines)-1) div 2;
 b := false;
 repeat begin
  if (_Splines[i].Min > x) then begin
   if (i = l) then break;
   u := i - 1;
   i := (u + l) div 2;
  end else if (_Splines[i].Max <= x) then begin
   if (i = u) then break;
   l := i + 1;
   i := (u + l) div 2;
  end else b := true;
 end until (b);
 _LastIndex := i;
 if (x < _Splines[i].Min) or (x > _Splines[i].Max) then exit;
 Result := _Splines[i].Fkt.Calc(x);
end;

function Polynom.TPolySpline.ToString : String;
var i : integer;
begin
 Result := '';
 for i := 0 to Length(_Splines) - 1 do Result := Result + '(' + FloatToStr(_Splines[i].Min) + ' | ' + FloatToStr(_Splines[i].Max) + ' | ' + _Splines[i].Fkt.ToString + ') ';
end;

function Polynom.TPolySpline.Derivate : TPolySpline;
var i : integer;
begin
 Result := TPolySpline.Create;
 SetLength(Result._Splines, Length(_Splines));
 for i := 0 to Length(_Splines)-1 do begin
  Result._Splines[i].Min := _Splines[i].Min;
  Result._Splines[i].Max := _Splines[i].Max;
  Result._Splines[i].Fkt := _Splines[i].Fkt.Derivate;
 end;
end;

function Polynom.TPolySpline.Integrate(c : Precision) : TPolySpline;
var i : integer;
begin
 Result := TPolySpline.Create;
 SetLength(Result._Splines, Length(_Splines));
 for i := 0 to Length(_Splines)-1 do begin
  Result._Splines[i].Min := _Splines[i].Min;
  Result._Splines[i].Max := _Splines[i].Max;
  Result._Splines[i].Fkt := _Splines[i].Fkt.Integrate(c);
  if (i > 0) then Result._Splines[i].Fkt[0] := Result._Splines[i-1].Fkt.Calc(Result._Splines[i].Min) - Result._Splines[i].Fkt.Calc(Result._Splines[i].Min);
 end;
end;

function Polynom.TPolySpline.ReadSpline(i : Cardinal) : TPolySplinePart; inline;
begin
 Result := _Splines[i];
end;

function Polynom.TPolySpline.ReadCount : Cardinal; inline;
begin
 Result := Length(_Splines);
end;

destructor Polynom.TPolySpline.Destroy;
var i : integer;
begin
 for i := Length(_Splines)-1 downto 0 do _Splines[i].Fkt.Free;
 SetLength(_Splines, 0);
 inherited;
end;




class function Polynom.NewtonInterpolation(const IP : TInterpolateInfo) : TPolynom;
var
  i, j, l : integer;
begin
 Result := HermiteInterpolation(IP); exit; // Newton interpolation not implemented!
 Result := TPolynom.Create;
 l := IP.Count-1;
 for i := 0 to l do begin
  //if (IP[l].derivation <> 0) then Exception.Create('Newton-Interpolation must not contain function values of derivations');

 end;
 Result[0] := 1;
 Result[2] := -1;
 Result[3] := 0.1;
end;

class function Polynom.Gauss(var M : TMatrix; var v : TVector; var b : TVector) : Boolean;
var
 d, i, j, k : integer;
 p : Precision;
begin
 Result := false;
 if (M.Rows <> M.Colums) then begin _LastError := 'Matrix ist not quadratic.'; exit; end;
 if (Length(v) <> M.Rows) or (Length(b) <> M.Rows) then begin _LastError := 'Vector dimension is not equal to Matrix dimension.'; exit; end;
 for d := 0 to M.Rows - 1 do begin // Obere Dreiecksmatrix erzeugen
  // Spaltenpivotsuche
  for i := d to M.Rows - 2 do begin
   k := i;
   for j := i to M.Rows - 1 do if (M[j, d]*M[j, d] > M[k, d]*M[k, d]) then k := j; // Zeilenindex mit Groesstem Element in Spalte d in k speichern
   M.SwitchRows(k, i);
   p := b[i];
   b[i] := b[k];
   b[k] := p;
  end;
  if (M[d,d] = 0) then begin _LastError := 'System cannot be solved due to linear dependency.'; exit; end;
  // Zeile unter d,d Nullen
  for i := d+1 to M.Rows - 1 do begin
   if (M[i, d] = 0) then break;
   b[i] := b[i] - b[d] / M[d, d] * M[i, d];
   for j := M.Colums-1 downto d do M[i, j] := M[i, j] - M[d, j] / M[d, d] * M[i, d];
  end;
 end;
 for d := M.Colums - 1 downto 0 do begin // Ergebnisvektor v berechnen
  p := 0;
  for i := d+1 to M.Colums-1 do p := p + v[i] * M[d, i];
  v[d] := (b[d] - p) / M[d, d];
 end;
 Result := true;
end;

class function Polynom.fak(v, n : Cardinal) : Cardinal;
begin
 Result := 1;
 for n := n downto 1 do begin
  Result := Result * v;
  if (v > 0) then v := v - 1;
 end;
end;

class function Polynom.HermiteInterpolation(const IP : TInterpolateInfo) : TPolynom;
var
  M : TMatrix;
  v, b : TVector;
  i, j : integer;
  x : Precision;
begin
 Result := nil;
 i := IP.count;
 SetLength(v, i);
 SetLength(b, i);
 M.Init(i, i, false);
 for i := 0 to IP.count-1 do begin
  x := 1;
  for j := 0 to IP.count-1 do begin
   M[i, j] := x * fak(j, IP[i].derivation);   // Hier: Ableitungen beachten!
   if (j >= IP[i].derivation) then x := x * IP[i].x;
  end;
  b[i] := IP[i].y;
 end;
 if (not Gauss(M, v, b)) then exit;
// ShowMessage(M.ToString + Char(13) + Char(10) + VecToStr(b) + Char(13) + Char(10) + VecToStr(v));
 Result := TPolynom.Create;
 for i := Length(v) - 1 downto 0 do Result[i] := v[i];
end;

class function Polynom.AdjustedSplinesOnFirstDerivation(var IP : TInterpolateInfo; AdjustSplines, AdjustPoints : Precision) : TPolySpline;
var
 n, _l, r, s   : integer;
 k, ku, ko     : Precision;
 g, _g, g_, gu, go : Precision;

 i, j, l, m : integer;
 mx, my     : Precision;

begin
 Result := TPolySpline.Create;
 IP.sort;
 _l := -1;
 s := 0;
 _g := -1;
 g := -1;
 for n := 0 to IP.count-1 do if (IP[n].derivation = 0) then begin
  ku := 0.5 - AdjustSplines;
  ko := 0.5 + AdjustSplines;
  gu := 1 - AdjustPoints;
  go := 1 + AdjustPoints;
  if (_g = -1) then _g := g;

  for r := 1 to trunc(AdjustSplines*20 + AdjustPoints * 20) + 1 do begin
   k := (ku + ko) / 2;
   g := (gu + go) / 2;
   for j := s to Length(Result._Splines)-1 do Result._Splines[j].Fkt.Free;
   SetLength(Result._Splines, s);
   j := -1;

   l := _l;
   for i := _l+1 to IP.count-1 do if (IP[i].derivation = 1) then begin
    if (l = -1) then begin l := i; continue; end;
    j := Length(Result._Splines);
    SetLength(Result._Splines, j+1);
    Result._Splines[j].Fkt := TPolynom.Create;
    mx := (IP[i].x + IP[l].x)/2;
    if (l = _l) then g_ := _g else g_ := g;
    if (IP[i].y*g - IP[l].y*g_ >= 0) then my := IP[l].y*g_ + (IP[i].y*g - IP[l].y*g_) * k
    else my := IP[l].y*g_ + (IP[i].y*g - IP[l].y*g_) * (1-k);
    Result._Splines[j].Fkt[3] := ((IP[i].x*(IP[l].y*g_ - my) + IP[i].y*g*(mx - IP[l].x) + IP[l].x*my - IP[l].y*g_*mx)/((IP[i].x*IP[i].x - IP[i].x*(IP[l].x + mx) + IP[l].x*mx)*(mx - IP[l].x)))/3;
    Result._Splines[j].Fkt[2] := ((IP[i].x*IP[i].x*(IP[l].y*g_ - my) + IP[i].y*g*(IP[l].x + mx)*(mx - IP[l].x) + IP[l].x*IP[l].x*my - IP[l].y*g_*mx*mx)/((IP[i].x*IP[i].x - IP[i].x*(IP[l].x + mx) + IP[l].x*mx)*(IP[l].x - mx)))/2;
    Result._Splines[j].Fkt[1] := (IP[i].x*IP[i].x*(IP[l].x*my - IP[l].y*g_*mx) + IP[i].x*(IP[l].y*g_*mx*mx - IP[l].x*IP[l].x*my) + IP[i].y*g*IP[l].x*mx*(IP[l].x - mx))/((IP[i].x*IP[i].x - IP[i].x*(IP[l].x + mx) + IP[l].x*mx)*(IP[l].x - mx));
    Result._Splines[j].Fkt[0] := 0;
    if (j = 0) then begin
     for m := 0 to IP.count-1 do if (IP[m].derivation = 0) then begin
      Result._Splines[0].Fkt[0] := - Result._Splines[0].Fkt.Calc(IP[m].x) + IP[m].y;
      break;
     end;
    end else Result._Splines[j].Fkt[0] := Result._Splines[j-1].Fkt.Calc(IP[l].x) - Result._Splines[j].Fkt.Calc(IP[l].x);
    Result._Splines[j].Min := IP[l].x;
    Result._Splines[j].Max := IP[i].x;
    l := i;
    if (IP[i].x >= IP[n].x) then break;
   end;

   if (j <= 0) then break;
   if (Result._Splines[j].Fkt.Calc(IP[n].x) > IP[n].y) then begin ko := k; go := g; end else begin ku := k; gu := g; end;
  end;

  s := Length(Result._Splines);
  _l := l;
  _g := g;
 end;
end;

class function Polynom.KubicSplines(var IP : TInterpolateInfo) : TPolySpline;
var
 Grad    : Array of Precision;
 P       : TPolynom;
 s       : Precision;
 i, j, n : integer;
begin
 Result := TPolySpline.Create;
 n := IP.count;
 if (n < 2) then exit;
 IP.sort;
 SetLength(Grad, n);
 for i := 1 to n-2 do if (IP[i-1].derivation <> 0) then Exception.Create('Kubic splines cannot be calculated for derivations!')
 //else if (IP[i].x = IP[i-1].x) or (IP[i].x = IP[i+1].x) then Exception.Create('Kubic splines cannot be calculated: Two identical x values!')
 else begin
  s := ((IP[i+1].y - IP[i].y)*(IP[i+1].y - IP[i].y)+(IP[i+1].x - IP[i].x)*(IP[i+1].x - IP[i].x)) / ((IP[i].y - IP[i-1].y)*(IP[i].y - IP[i-1].y) + (IP[i].x - IP[i-1].x)*(IP[i].x - IP[i-1].x) + (IP[i+1].y - IP[i].y)*(IP[i+1].y - IP[i].y)+(IP[i+1].x - IP[i].x)*(IP[i+1].x - IP[i].x));
  Grad[i] := (IP[i].y - IP[i-1].y)/(IP[i].x - IP[i-1].x)*s + ((IP[i+1].y - IP[i].y)/(IP[i+1].x - IP[i].x))*(1-s);
 end;
 Grad[0] := (IP[1].y - IP[0].y)/(IP[1].x - IP[0].x);
 Grad[n-1] := (IP[n-1].y - IP[n-2].y)/(IP[n-1].x - IP[n-2].x);
 for i := 0 to n-2 do begin
  P := TPolynom.Create;
  P[3] := (Grad[i]*(IP[i].x - IP[i+1].x) + Grad[i+1]*(IP[i].x - IP[i+1].x) - 2*(IP[i].y - IP[i+1].y))/(IP[i].x*IP[i].x*IP[i].x - 3*IP[i].x*IP[i].x*IP[i+1].x + 3*IP[i].x*IP[i+1].x*IP[i+1].x - IP[i+1].x*IP[i+1].x*IP[i+1].x);
  P[2] := (Grad[i]*(IP[i].x*IP[i].x + IP[i].x*IP[i+1].x - 2*IP[i+1].x*IP[i+1].x) + Grad[i+1]*(2*IP[i].x*IP[i].x - IP[i].x*IP[i+1].x - IP[i+1].x*IP[i+1].x) - 3*(IP[i].y - IP[i+1].y)*(IP[i].x + IP[i+1].x))/((IP[i].x*IP[i].x - 2*IP[i].x*IP[i+1].x + IP[i+1].x*IP[i+1].x)*(IP[i+1].x - IP[i].x));
  P[1] := (Grad[i]*IP[i+1].x*(2*IP[i].x*IP[i].x - IP[i].x*IP[i+1].x - IP[i+1].x*IP[i+1].x) + IP[i].x*(Grad[i+1]*(IP[i].x*IP[i].x + IP[i].x*IP[i+1].x - 2*IP[i+1].x*IP[i+1].x) - 6*IP[i+1].x*(IP[i].y - IP[i+1].y)))/((IP[i].x - IP[i+1].x)*(IP[i].x*IP[i].x - 2*IP[i].x*IP[i+1].x + IP[i+1].x*IP[i+1].x));
  P[0] := (Grad[i]*IP[i].x*IP[i+1].x*IP[i+1].x*(IP[i].x - IP[i+1].x) + Grad[i+1]*IP[i].x*IP[i].x*IP[i+1].x*(IP[i].x - IP[i+1].x) + IP[i].y*IP[i+1].x*IP[i+1].x*(IP[i+1].x - 3*IP[i].x) + IP[i+1].y*IP[i].x*IP[i].x*(3*IP[i+1].x - IP[i].x))/((IP[i].x*IP[i].x - 2*IP[i].x*IP[i+1].x + IP[i+1].x*IP[i+1].x)*(IP[i+1].x - IP[i].x));
  j := Length(Result._Splines);
  SetLength(Result._Splines, j+1);
  Result._Splines[j].Fkt := P;
  Result._Splines[j].Min := IP[i].x;
  Result._Splines[j].Max := IP[i+1].x;
 end;
end;

class function Polynom.KubicSplines2(var IP : TInterpolateInfo; hermiteRB : Boolean; fa, fb : Precision) : TPolySpline;
type
 row = Array [0..3] of Precision; // [0] : Diagonalelement  [1] : Wert hinter Diagonalelement (Ausnahme letzte Zeile, Hermite RB)  [2] : m (Lösung)  [3] : Ergebnisvektorelement
var
 m : Array of row;
 c,d,f : precision;
 P : TPolynom;
 i, j : integer;
begin
 Result := nil;
 i := IP.count;
 if (i < 2) then begin _LastError := 'Kubic Splines need at least two points!'; exit; end;
 SetLength(m, i+1);
 m[i][2] := 0;
 i := i-1; // i = Length(IP)-1
 IP.sort;
 if (IP[0].derivation <> 0) or (IP[i].derivation <> 0) then begin _LastError := 'Kubic Splines cannot be calculated for derivations!'; exit; end;
 // Randbedingungen
 if (hermiteRB) then begin
  f := (IP[1].x - IP[0].x);
  m[0][0] := f / 3; m[0][1] := f / 6; if (f <> 0) then m[0][3] := (IP[1].y - IP[0].y) / f - fa;
  f := (IP[i].x - IP[i-1].x);
  m[i][0] := f / 3; m[i][1] := f / 6; if (f <> 0) then m[i][3] := fb - (IP[i].y - IP[i-1].y) / f;
 end else begin
  m[0][0] := 1; m[0][1] := 0; m[0][3] := 0;
  m[i][0] := 1; m[i][1] := 0; m[i][3] := 0;
 end;
 // Obere (diagonaldominante) Dreiecksmatrix in m speichern
 i := i-1; // i = Length(IP)-2
 for j := 1 to i do begin
  if (IP[j].x = IP[j-1].x) or (IP[j].x = IP[j+1].x) then begin _LastError := 'Equal x-values of two points!'; exit; end
  else if (IP[j].derivation <> 0) then begin _LastError := 'Kubic Splines cannot be calculated for derivations!'; exit; end;
  f := (IP[j].x - IP[j-1].x) / (m[j-1][0] * 6);
  m[j][0] := (IP[j+1].x - IP[j-1].x) / 3 - f * m[j-1][1];
  m[j][1] := (IP[j+1].x - IP[j].x) / 6;
  m[j][3] := (IP[j+1].y - IP[j].y) / (IP[j+1].x - IP[j].x) - (IP[j].y - IP[j-1].y) / (IP[j].x - IP[j-1].x) - f * m[j-1][3];
 end;
 i := IP.count-1;
 if (hermiteRB) then begin
  f := (m[i][1]) / m[i-1][0];
  m[i][0] := m[i][0] - f * m[i-1][1];
  m[i][1] := 0;
  m[i][3] := m[i][3] - f * m[i-1][3];
 end;
 // m berechnen
 for j := i downto 0 do m[j][2] := (m[j][3] - m[j+1][2] * m[j][1]) / m[j][0];
 Result := TPolySpline.Create;
 SetLength(Result._Splines, i); i := i-1;
 for j := 0 to i do begin
  f := (IP[j+1].x - IP[j].x);
  c := (IP[j+1].y - IP[j].y) / f - f / 6 * (m[j+1][2] - m[j][2]);
  d := IP[j].y - f*f/6 * m[j][2];
  P := TPolynom.Create;
  P[3] := (m[j+1][2] - m[j][2]) / (6*f);
  P[2] := (m[j][2] * IP[j+1].x - m[j+1][2] * IP[j].x) / (2*f);
  P[1] := (m[j][2] * IP[j+1].x * IP[j+1].x - m[j+1][2] * IP[j].x * IP[j].x - 2*c*f) / (-2*f);
  P[0] := (m[j][2] * IP[j+1].x * IP[j+1].x * IP[j+1].x - m[j+1][2] * IP[j].x * IP[j].x * IP[j].x - 6*f*(c*IP[j].x - d)) / (6*f);
  Result._Splines[j].Fkt := P;
  Result._Splines[j].Min := IP[j].x;
  Result._Splines[j].Max := IP[j+1].x;
 end;
end;

class function Polynom.LinearSplines(var IP : TInterpolateInfo) : TPolySpline;
var
 P : TPolynom;
 i, j : integer;
begin
 Result := TPolySpline.Create;
 i := IP.count-1;
 if (i < 0) then exit;
 if (IP[i].derivation <> 0) then begin Result.Free; Result := nil; _LastError := 'Linear Splines cannot be calculated for derivations!'; exit; end;
 SetLength(Result._Splines, i);
 FillDWord(Result._Splines[0], i * SizeOf(Result._Splines[0]) div 4, 0);
 IP.sort;
 i := i-1;
 for j := 0 to i do begin
  if (IP[j+1].x = IP[j].x) then begin Result.Free; Result := nil; _LastError := 'Two points got the same x-value!'; exit; end
  else if (IP[j].derivation <> 0) then begin Result.Free; Result := nil; _LastError := 'Linear Splines cannot be calculated for derivations!'; exit; end;
  P := TPolynom.Create;
  P[1] := (IP[j+1].y - IP[j].y) / (IP[j+1].x - IP[j].x);
  P[0] := IP[j].y - IP[j].x * P[1];
  Result._Splines[j].Min := IP[j].x;
  Result._Splines[j].Max := IP[j+1].x;
  Result._Splines[j].Fkt := P;
 end;
end;




// Bugfixes FPC generics

class function Polynom.IsNaN(v : Precision) : Boolean;
begin
 case (PTypeInfo(TypeInfo(Precision))^.Kind) of
   tkFloat : begin
     case (GetTypeData(PTypeInfo(TypeInfo(Precision)))^.FloatType) of
      ftSingle   : Result := PCardinal(@v)^ = $FFFFFFFF;
      ftDouble   : Result := (PCardinal(@v)^ = $FFFFFFFF) and (PCardinal(@v+4)^ = $FFFFFFFF);
      ftExtended : Result := (PCardinal(@v)^ = $FFFFFFFF) and (PCardinal(@v+4)^ = $FFFFFFFF) and (PWord(@v+8)^ = $FFFF);
     end;
   end;
  end;
end;

class function Polynom.NaN : Precision;
begin
 case (PTypeInfo(TypeInfo(Precision))^.Kind) of
   tkFloat : begin
     case (GetTypeData(PTypeInfo(TypeInfo(Precision)))^.FloatType) of
      ftSingle   : PCardinal(@Result)^ := $FFFFFFFF;
      ftDouble   : begin PCardinal(@Result)^ := $FFFFFFFF; PCardinal(@Result+4)^ := $FFFFFFFF; end;
      ftExtended : begin PCardinal(@Result)^ := $FFFFFFFF; PCardinal(@Result+4)^ := $FFFFFFFF; PWord(@Result+8)^ := $FFFF; end;
     end;
   end;
  end;
end;

class function Polynom.FloatToStr(f : Precision) : String; inline;
begin
 Result := '?';
 case (PTypeInfo(TypeInfo(Precision))^.Kind) of
  tkFloat : begin
    case (GetTypeData(PTypeInfo(TypeInfo(Precision)))^.FloatType) of
     ftSingle   : Result := SysUtils.FloatToStr(Single(f));
     ftDouble   : Result := SysUtils.FloatToStr(Double(f));
     ftExtended : Result := SysUtils.FloatToStr(Extended(f));
    end;
  end;
 end;
end;

class function Polynom.Trunc(f : Precision) : Integer; inline;
begin
 case (PTypeInfo(TypeInfo(Precision))^.Kind) of
  tkFloat : begin
    case (GetTypeData(PTypeInfo(TypeInfo(Precision)))^.FloatType) of
     ftSingle   : Result := Math.Floor(Single(f));
     ftDouble   : Result := Math.Floor(Double(f));
     ftExtended : Result := Math.Floor(Extended(f));
    end;
  end;
 end;
end;

class function Polynom.StrToIntErr(s : String; i : integer) : integer;
begin
 try
  Result := StrToInt(s);
 except
  Result := i;
 end;
end;

class function Polynom.StrToFloatErr(s : String; f : float) : float;
begin
 try
  Result := StrToFloat(s);
 except
  Result := f;
 end;
end;

end.


