unit functions;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, PolynomUtil, Graph2D, settings, Math, dialogs;
type
 TSRCValues = class
  const
   Tinfinity : Precision = 10E6; // this value should be larger than the period duration
  type
   procedureOfObject = procedure of Object;
   fktOfObject = function(x:Precision):Precision of object;
   TTransistorState = (blocking, diode, channel);
   TSwitchAction = (Q1_open, Q1_close, Q2_open, Q2_close);
   TState = object
    private
     procedure writeInvalid (v : Boolean);
     function readInvalid : Boolean; inline;
    public
     Q1, Q2 : TTransistorState;
     D : Boolean;
     property invalid : Boolean read readInvalid write writeInvalid;
     function toString : String;
   end;
   TPFunction = class(Poly.IFunction)
    strict private
     pf : fktOfObject;
     pOffset, pEndTime : PPrecision;
    public
     constructor Create(_pf : fktOfObject; _pOffset, _pEndTime : PPrecision);
     function Calc(x : Precision) : Precision; override;
   end;
  strict private
   function readDuration : Precision; inline;
   procedure writeDuration(v : Precision); inline;
   // circuit
   class procedure initVal;
  protected
   endTime : Precision;
   nextState : TState;
   _state : TState;
   function index : integer; inline;
   procedure doUpdate(prev : TSRCValues); virtual;
   function fCalcI1 (t : Precision) : Precision;
   function fCalcIh (t : Precision) : Precision; virtual; abstract;
   function fCalcI2 (t : Precision) : Precision; virtual; abstract;
   function fCalcVLh (t : Precision) : Precision; virtual; abstract;
   function fCalcVQ2 (t : Precision) : Precision; virtual; abstract;
   function fCalcVcres (t : Precision) : Precision; virtual; abstract;
   function fCalcIsec (t : Precision) : Precision;
   // circuit
   class function sig (v : Precision) : Precision; inline;
   class function checkState(n : integer; state : TState) : Boolean;
   class function getNextSwitchAction(tOffset : Precision; var tEndTime : Precision) : TSwitchAction;
   class procedure log (v : String); inline;
  public
   calcI1, calcIh, calcI2, calcVLh, calcVq2, calcVcres, calcIsec : TGraph2D.TFunction;
   Vcres_0, Vq2_0, Ih_0, I2_0 : Precision;
   DiodesForward : Boolean;
   t_offset : Precision;
   durationMarker : TGraph2D.TBar;
   constructor Create(__state : TState);
   property duration : Precision read readDuration write writeDuration;
   property state : TState read _state;
   destructor Destroy; override;
   // circuit
   class var
    Vin, Vout, Vcres_0_def, Ih_0_def, I2_0_def : Precision;
    Cq1, Cq2, Cres, Ls1, Lh, Ls2, R2, U_ : Precision;
    Ih_charge, I2_charge : Precision;
    switchOnVoltage, switchOffCurrent, switchOnCurrent : Precision;
    Ton, Toff : Precision;
    isInitVal : Boolean;
    period : Array of TSRCValues;
    lastError : String;
    logArr : Array of String;
    logEnabled : Boolean;
   class procedure init;
   class procedure clear;
   class function update : Boolean;
   class function gotoSteadyState(charge : Precision; voltage : Precision) : Precision;
 end;

 TDOpen = class(TSRCValues)
  type
   TParamSet = Array [0..1] of Precision;
   TApproxFkt = class (Poly.IFunction)
    private
     params : TParamSet;
     owner : TDOpen;
    public
     function Calc(t : Precision) : Precision; override;
   end;
  strict private
   a, b, omega : Precision;
   function derivate(const params : TParamSet) : TParamSet; inline;
   function integrate(const params : TParamSet) : TParamSet; inline;
   function calc(const params : TParamSet; t : Precision) : Precision; inline;
   function par(pa, pb : Precision) : TParamSet; inline;
   function getFirstZero(params : TParamSet; pd : Precision; gt : Boolean) : Precision;
  protected
   procedure doUpdate(prev : TSRCValues); override;
   function fCalcIh (t : Precision) : Precision; override;
   function fCalcI2 (t : Precision) : Precision; override;
   function fCalcVLh (t : Precision) : Precision; override;
   function fCalcVq2 (t : Precision) : Precision; override;
   function fCalcVcres (t : Precision) : Precision; override;
 end;

 TDShort = class(TSRCValues)
  type
   TParamSet = Array [0..2] of Precision;
   TApproxFkt = class (Poly.IFunction)
    private
     params : TParamSet;
     owner : TDShort;
    public
     function Calc(t : Precision) : Precision; override;
   end;
  strict private
   _A, _B, _C, _D : Precision;
   p, q, u, v : Precision;
   a, a_, b, b_, c, c_, d, d_ : Precision;
   omega, beta1, beta2 : Precision;
   function derivate(const params : TParamSet) : TParamSet; inline;
   function integrate(const params : TParamSet) : TParamSet; inline;
   function calc(const params : TParamSet; t : Precision) : Precision; inline;
   function par(pa, pb, pc : Precision) : TParamSet; inline;
   function getFirstZero(params : TParamSet; pd : Precision; gt : Boolean) : Precision;
  protected
   procedure doUpdate(prev : TSRCValues); override;
   function fCalcIh (t : Precision) : Precision; override;
   function fCalcI2 (t : Precision) : Precision; override;
   function fCalcVLh (t : Precision) : Precision; override;
   function fCalcVq2 (t : Precision) : Precision; override;
   function fCalcVcres (t : Precision) : Precision; override;
 end;

implementation

procedure TSRCValues.TState.writeInvalid (v : Boolean);
begin
 if (v) then begin Q1 := channel; Q2 := channel; end
 else begin Q1 := blocking; Q2 := blocking; end;
 D := false;
end;

function TSRCValues.TState.readInvalid : Boolean;
begin
 Result := (Q1 <> blocking) and (Q2 <> blocking);
end;

function TSRCValues.TState.toString : String;
begin
 if (Q1 <> blocking) and (Q2 <> blocking) then begin
  Result := 'i';
  exit;
 end;
 if (Q1 <> blocking) then begin
  if (Q1 = diode) then Result := 'Qd1'
  else Result := 'Qc1';
 end else if (Q2 <> blocking) then begin
  if (Q2 = diode) then Result := 'Qd2'
  else Result := 'Qc2';
 end else Result := '0';
 if (D) then Result := Result + '_D';
end;






class procedure TSRCValues.init;
begin
 period := nil;
 lastError := '';
 logArr := nil;
 logEnabled := false;
 initVal;
end;

class procedure TSRCValues.clear;
var i : integer;
begin
 SetLength(logArr, 0);
 for i := 0 to Length(period)-1 do period[i].Destroy;
 SetLength(period, 0);
end;

class procedure TSRCValues.log (v : String); inline;
begin
 if (not logEnabled) then exit;
 SetLength(LogArr, Length(LogArr)+1);
 LogArr[Length(LogArr)-1] := v;
end;

class procedure TSRCValues.initVal;
begin
 log('initial values');
 Vin := 0;
 Vout := 0;
 Vcres_0_def := 0;
 Ih_0_def := 0;
 I2_0_def := 0;
 Cq1 := 1;
 Cq2 := 1;
 Cres := 1;
 Ls1 := 1;
 Lh := 1;
 Ls2 := 1;
 R2 := 1;
 Ton := 0.1;
 Toff := 0.1;
 isInitVal := true;
 I2_charge := 0;
 Ih_charge := 0;
 switchOnVoltage := 0;
 switchOnCurrent := 0;
 switchOffCurrent := 0;
end;

class function TSRCValues.getNextSwitchAction(tOffset : Precision; var tEndTime : Precision) : TSwitchAction;
begin
 tEndTime := 0;
 repeat begin
  if          (tEndTime > tOffset) then begin
   Result := Q1_close;
   break;
  end;
  tEndTime := tEndTime + Ton;
  if (tEndTime > tOffset) then begin
   Result := Q1_open;
   break;
  end;
  tEndTime := tEndTime + Toff;
  if (tEndTime > tOffset) then begin
   Result := Q2_close;
   break;
  end;
  tEndTime := tEndTime + Ton;
  if (tEndTime > tOffset) then begin
   Result := Q2_open;
   break;
  end;
  tEndTime := tEndTime + Toff; // Additiv iterieren, wegen Rechengenauigkeit -.-
 end until false;
 if (logEnabled) then begin
   if (Result = Q1_close) then log('   next switch action: Q1_close    t=' + floattostr(tEndTime));
   if (Result = Q1_open) then log('   next switch action: Q1_open    t=' + floattostr(tEndTime));
   if (Result = Q2_close) then log('   next switch action: Q2_close    t=' + floattostr(tEndTime));
   if (Result = Q2_open) then log('   next switch action: Q2_open    t=' + floattostr(tEndTime));
 end;
end;

class function TSRCValues.checkState(n : integer; state : TState) : Boolean;
begin
 Result := true;
 if (n >= Length(period)) then begin
  SetLength(period, n+1);
  period[n] := nil;
 end else if (period[n] <> nil) and ((period[n].state.Q1 <> state.Q1) or (period[n].state.Q2 <> state.Q2) or (period[n].state.D <> state.D)) then begin
  period[n].Destroy;
  period[n] := nil;
 end;
 if (state.invalid) then begin Result := false; exit; end; // Ungültiger State
 if (period[n] = nil) then if (state.D) then period[n] := TDShort.create(state)
 else period[n] := TDOpen.Create(state);
end;

class function TSRCValues.update : Boolean;
var
  i,j : integer;
  d, d_ : Precision;
  startState : TState;
begin
 log(' period   Vin=' + floattostr(Vin) + '   Vout=' + floattostr(Vout));
 LastError := '';
 Result := true;
 I2_charge := 0;
 Ih_charge := 0;
 switchOnVoltage := 0;
 switchOnCurrent := 0;
 switchOffCurrent := 0;
 if (Poly.IsNaN(Vin)) or (Vin < 0) then InitVal;
 if (Poly.IsNaN(Vout)) or (Vout < 0) then InitVal;
 if (Poly.IsNaN(Vcres_0_def)) then InitVal;
 if (Poly.IsNaN(Ih_0_def)) then InitVal;
 if (Poly.IsNaN(Cq1)) or (Cq1 <= 0) then InitVal;
 if (Poly.IsNaN(Cq2)) or (Cq2 <= 0) then InitVal;
 if (Poly.IsNaN(Cres)) or (Cres <= 0) then InitVal;
 if (Poly.IsNaN(Ls1)) or (Ls1 <= 0) then InitVal;
 if (Poly.IsNaN(Ls2)) or (Ls2 <= 0) then InitVal;
 if (Poly.IsNaN(Lh)) or (Lh <= 0) then InitVal;
 if (Poly.IsNaN(R2)) or (R2 <= 0) then InitVal;
 if (Poly.IsNaN(Ton)) or (Ton <= 0) then InitVal;
 if (Poly.IsNaN(Toff)) or (Toff <= 0) then InitVal;
 startState.Q1 := channel;
 startState.Q2 := blocking;
 startState.D := I2_0_def <> 0;
 checkState(0, startState);
 period[0].doUpdate(nil);
 i := 1;
 d := period[0].duration;
 d_ := 0;
 j := 0;
 while (d < Ton+Toff+Ton+Toff) do begin
  if (not checkState(i, period[i-1].nextState)) then begin
   period[i] := nil;
   Result := false;
   lastError := 'A state ('+period[i-1].state.toString+', @Vcres_0 = ' + FloatToSTr(period[i-1].Vcres_0) + ') has no follower (nextState = none)';
   break;
  end;
  period[i].doUpdate(period[i-1]);
  d := period[i].t_offset + period[i].duration;
  if (d = d_) then j := j + 1
  else j := 0;
  if (j > 4) then begin
   Result := false;
   lastError := 'Zero Duration Loop detected!';
   break;
  end;
  d_ := d;
  i := i + 1;
  if (i > 100) then begin lastError := 'Too many states!'; break; end;
 end;
 for j := Length(period)-1 downto i do period[j].Free;
 SetLength(period, i);
end;

class function TSRCValues.gotoSteadyState(charge : Precision; voltage : Precision) : Precision;
var
 V, C, lastC : Precision;
 d, lastD : Precision;
 n, i, j, cnt : integer;
 lastCharge : Precision;
begin
 log('goto steady state');
 V := Vin/2;
 d := V / 5;
 lastd := d;
 i := 0;
 j := 0;
 if (voltage >= 0) then cnt := 30 else cnt := 80;
 if (charge >= 0) and (voltage >= 0) then ShowMessage('Not Implemented');
 repeat begin
  if (voltage >= 0) then begin
   Vout := voltage;
   i := i+1;
   if (not update) then break;
  end else begin
   Vout := V;
   if (not update) then break;
   C := I2_charge;
   if (C = 0) then begin
    d := - Abs(lastD)/2;
    j := j + 1;
   end else begin
    if (C > Charge) then d := Vin/4 * 0.8**i else d := - Vin/4 * 0.8**i;
    lastD := d;
    i := i + 1;
   end;
   //showmessage('Ih_charge = '+floattostr(Ih_charge) + '    I2_charge = '+floattostr(C) + ' - ' + floattostr(charge) + ' ==> ' + floattostr(d));
   V := V + d;
  end;
  n := Length(period)-1;
  // set latest values of last period as init values for next period
  I2_0_def := period[n].fCalcI2(period[n].duration);
  Ih_0_def := period[n].fCalcIh(period[n].duration) - (Ih_charge)/(2*(Ton+Toff));
  Vcres_0_def := period[n].fCalcVcres(period[n].duration);
  if (v < 0) then break;
  if (j > cnt*2) then begin
   LastError := 'Cannot find stationary state!';
   break;
  end;
 end until (i > cnt);
 //Showmessage(floattostr(I2_charge) + '   ' + floattostr(charge));
 Result := v;
end;

class function TSRCValues.sig (v : Precision) : Precision; inline;
begin
 if (v < 0) then Result := -1
 else Result := 1;
end;





constructor TSRCValues.Create(__state : TState);
begin
 inherited Create;
 _state := __state;
 calcI1 := TGraph2D.TFunction.Create(TGraph2D(Graph), TPFunction.Create(@fCalcI1, @t_offset, @endTime));
 calcI1.Name := 'i_1';
 calcIh := TGraph2D.TFunction.Create(TGraph2D(Graph), TPFunction.Create(@fCalcIh, @t_offset, @endTime));
 calcIh.Name := 'i_h';
 calcI2 := TGraph2D.TFunction.Create(TGraph2D(Graph), TPFunction.Create(@fCalcI2, @t_offset, @endTime));
 calcI2.Name := 'i_2*';
 calcVLh := TGraph2D.TFunction.Create(TGraph2D(Graph), TPFunction.Create(@fCalcVLh, @t_offset, @endTime));
 calcVLh.Name := 'v_Lh';
 calcVq2 := TGraph2D.TFunction.Create(TGraph2D(Graph), TPFunction.Create(@fCalcVq2, @t_offset, @endTime));
 calcVq2.Name := 'v_mid';
 calcVcres := TGraph2D.TFunction.Create(TGraph2D(Graph), TPFunction.Create(@fCalcVcres, @t_offset, @endTime));
 calcVcres.Name := 'v_Cres';
 calcIsec := TGraph2D.TFunction.Create(TGraph2D(Graph), TPFunction.Create(@fCalcIsec, @t_offset, @endTime));
 calcIsec.Name := 'i_sec';
 TGraph2D(Graph).XAxis[0].Name := 't';
 calcVLh.YAxis := TGraph2D(Graph).YAxis[1];
 calcVq2.YAxis := TGraph2D(Graph).YAxis[1];
 calcVcres.YAxis := TGraph2D(Graph).YAxis[1];
 calcIh.Style := DASHED;
 calcI2.Style := DOTTED;
 calcVLh.Style := DASHED;
 calcVcres.Style := DOTTED;
 durationMarker := TGraph2D.TBar.Create(TGraph2D(Graph), VERTICAL);
 nextState.invalid := true;
 endTime := 0;
end;

function TSRCValues.index : integer;
begin
 for Result := 0 to Length(period) - 1 do if (period[Result] = self) then exit;
 Result := -1;
end;

function TSRCValues.fCalcI1 (t : Precision) : Precision;
begin
 Result := fCalcIh(t) + fCalcI2(t);
end;

function TSRCValues.fCalcIsec (t : Precision) : Precision;
begin
 Result := abs(fCalcI2(t)) * U_;
end;

procedure TSRCValues.doUpdate(prev : TSRCValues);
var activeSwitch : Boolean;
begin
 nextState.invalid := true;
 if (prev = nil) then begin
  Ih_0 := Ih_0_def;
  I2_0 := I2_0_def;
  VCres_0 := VCres_0_def;
  Vq2_0 := Vin;
  DiodesForward := true;
  t_offset := 0;
 end else begin
  Ih_0 := prev.fCalcIh(prev.duration);
  I2_0 := prev.fCalcI2(prev.duration);
  VCres_0 := prev.fCalcVcres(prev.duration);
  Vq2_0 := prev.fCalcVq2(prev.duration);
  t_offset := prev.endTime;
  if (prev.InheritsFrom(TDShort)) then DiodesForward := prev.DiodesForward
  else DiodesForward := prev.fCalcVLh(prev.duration) >= 0;
 end;
 activeSwitch := false;
 if (state.Q1 = channel) and ((prev = nil) or (prev.state.Q1 <> channel)) then begin // Einschaltmoment Q1
  if (Abs(Vq2_0 - Vin) > switchOnVoltage) then switchOnVoltage := Abs(Vq2_0 - Vin);
  if (Abs(Ih_0) > switchOnCurrent) then switchOnCurrent := Abs(Ih_0);
 end;
 if (state.Q2 = channel) and ((prev = nil) or (prev.state.Q2 <> channel)) then begin // Einschaltmoment Q2
  if (Abs(Vq2_0 - 0) > switchOnVoltage) then switchOnVoltage := Abs(Vq2_0 - 0);
  if (Abs(Ih_0 + I2_0) > switchOnCurrent) then switchOnCurrent := Abs(Ih_0 + I2_0);
 end;
 if (state.Q1 <> channel) and (prev <> nil) and (prev.state.Q1 = channel) or (state.Q2 <> channel) and (prev <> nil) and (prev.state.Q2 = channel) then begin // Abschaltmoment Q1 oder Q2
  if (Abs(Ih_0 + I2_0) > switchOffCurrent) then switchOffCurrent := Abs(Ih_0 + I2_0);
 end;
 if (state.Q1 <> blocking) then Vq2_0 := Vin;
 if (state.Q2 <> blocking) then Vq2_0 := 0;
 log('  state='+ state.toString + ' from t='+floattostr(t_offset) + '   Vcres,0=' + floattostr(Vcres_0) + '   Vq2,0=' + floattostr(Vq2_0) + '   Ih,0=' + floattostr(Ih_0) + '   I2,0=' + floattostr(I2_0));
end;

function TSRCValues.readDuration : Precision; inline;
begin
 Result := endTime - t_offset;
end;

procedure TSRCValues.writeDuration(v : Precision); inline;
begin
 endTime := t_offset + v;
end;

destructor TSRCValues.Destroy;
begin
 calcI1.Free;
 calcIh.Free;
 calcI2.Free;
 calcVLh.Free;
 calcVq2.Free;
 calcVcres.Free;
 calcIsec.Free;
 durationMarker.free;
 inherited;
end;




constructor TSRCValues.TPFunction.Create(_pf : fktOfObject; _pOffset, _pEndTime : PPrecision);
begin
 inherited Create;
 pf := _pf;
 pOffset := _pOffset;
 pEndTime := _pEndTime;
end;

function TSRCValues.TPFunction.Calc(x : Precision) : Precision;
begin
 if (x < pOffset^) or (x > pEndTime^) then Result := Poly.NaN
 else begin
  x := x - pOffset^;
  Result := pf(x);
 end;
end;





function TDOpen.TApproxFkt.Calc(t : Precision) : Precision;
begin
 Result := owner.calc(params, t);
end;


procedure TDOpen.doUpdate(prev : TSRCValues);
var
 Vmid : Precision;
 Cr   : Precision;
 d1, d2, d3 : Precision;
 action     : TSwitchAction;
 _p : TParamSet;
begin
 inherited;
 // Case dependant values
 if (state.Q1 <> blocking) then begin Vmid := Vin; Cr := Cres; end
 else if (state.Q2 <> blocking) then begin Vmid := 0; Cr := Cres; end
 else begin Vmid := Vq2_0; Cr := 1/(1/Cres + 1/(Cq1+Cq2)); end;
 // differential equation parameters
 b := Ih_0;
 a := (Vmid - Vcres_0) * sqrt(Cr/(Lh+Ls1));
 omega := 1/sqrt(Cr*(Lh+Ls1));
 // duration
 action := GetNextSwitchAction(t_offset, endTime);
 _p := derivate(par(a, b));
 d1 := getFirstZero(_p, Vout/Lh, true);
 d2 := getFirstZero(_p, -Vout/Lh, false);
 if (d2 < d1) then d1 := d2;
 if (state.Q1 = channel) then begin
  if (action = Q1_open) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := false; end;
  if (d1 < duration) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := true; duration := d1 end;
 end else if (state.Q2 = channel) then begin
  if (action = Q2_open) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := false; end;
  if (d1 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := channel; nextState.D := true; duration := d1 end;
 end else if (state.Q1 = diode) then begin
  d2 := getFirstZero(par(a, b), 0, true);
  if (action = Q1_open) then begin nextState.Q1 := diode; nextState.Q2 := blocking; nextState.D := false; end;
  if (action = Q1_close) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := false; end;
  if (action = Q2_close) then begin nextState.Q2 := channel; nextState.Q1 := blocking; nextState.D := false; end;
  if (d2 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := false; duration := d2 end;
  if (d1 < duration) then begin nextState.Q1 := diode; nextState.Q2 := blocking; nextState.D := true; duration := d1 end;
 end else if (state.Q2 = diode) then begin
  d2 := getFirstZero(par(a, b), 0, false);
  if (action = Q2_open) then begin nextState.Q1 := blocking; nextState.Q2 := diode; nextState.D := false; end;
  if (action = Q2_close) then begin nextState.Q1 := blocking; nextState.Q2 := channel; nextState.D := false; end;
  if (action = Q1_close) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := false; end;
  if (d2 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := false; duration := d2 end;
  if (d1 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := diode; nextState.D := true; duration := d1 end;
 end else begin
  _p := integrate(par(a, b));
  d2 := getFirstZero(_p, (Vq2_0 - 0) * (Cq1+Cq2) + _p[1], true);
  d3 := getFirstZero(_p, (Vq2_0 - Vin) * (Cq1+Cq2) + _p[1], false);
  if (action = Q1_close) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := false; end;
  if (action = Q2_close) then begin nextState.Q1 := blocking; nextState.Q2 := channel; nextState.D := false; end;
  if (d2 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := diode; nextState.D := false; duration := d2; end;
  if (d3 < duration) then begin nextState.Q1 := diode; nextState.Q2 := blocking; nextState.D := false; duration := d3; end;
  if (d1 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := true; duration := d1; end;
 end;
 durationMarker.Position := duration + t_offset;
 log('   next state: '+nextState.toString+'    d=' + floattostr(duration));
 // charge
 _p := integrate(par(a, b));
 Ih_charge := Ih_charge + (calc(_p, duration) - _p[1]);
end;

function TDOpen.getFirstZero(params : TParamSet; pd : Precision; gt : Boolean) : Precision;
var
 f, df : TApproxFkt;
 i : integer;
 s : Precision;
begin
 if (params[1] > pd+Vin/10000) and (gt) or (params[1] < pd-Vin/10000) and (not gt) then begin Result := 0; exit; end; // ist bereits zu Beginn erreicht
 if (sqrt(params[0]**2 + params[1]**2) < abs(pd)) then begin Result := Tinfinity; exit; end; // wird nie erreicht
 f := TApproxFkt.Create;
 f.Owner := self;
 f.params := params;
 df := TApproxFkt.Create;
 df.Owner := self;
 df.params := derivate(params);
 i := 0;
 if (gt) then s := 1 else s := -1;
 while (i < 4) do begin
  Result := i * PI/(2*omega) + 1E-20;
  if (f.Newton(pd, df, Result, i * PI/(2*omega), (i+1) * PI/(2*omega), s)) then break;
  i := i+1;
 end;
 if (i = 4) then Result := Tinfinity;
 f.Destroy;
 df.Destroy;
end;

function TDOpen.derivate(const params : TParamSet) : TParamSet;
begin
 Result[0] := -params[1]*omega;
 Result[1] := params[0]*omega;
end;

function TDOpen.integrate(const params : TParamSet) : TParamSet;
begin
 Result[0] := params[1]/omega;
 Result[1] := -params[0]/omega;
end;

function TDOpen.calc(const params : TParamSet; t : Precision) : Precision;
begin
 Result := params[0]*sin(omega*t) + params[1]*cos(omega*t);
end;

function TDOpen.par(pa, pb : Precision) : TParamSet;
begin
 Result[0] := pa;
 Result[1] := pb;
end;

function TDOpen.fCalcVq2(t : precision) : precision;
var _p : TParamSet;
begin
 if (state.Q1 <> blocking) then Result := Vin
 else if (state.Q2 <> blocking) then Result := 0
 else begin
  _p := integrate(par(a, b));
  Result := Vq2_0 - (calc(_p, t) - _p[1]) / (Cq1+Cq2);
 end;
end;

function TDOpen.fCalcVLh(t : precision) : precision;
begin
 Result := Lh*calc(derivate(par(a,b)), t);
end;

function TDOpen.fCalcVcres (t : Precision) : Precision;
var _p : TParamSet;
begin
 _p := integrate(par(a, b));
 Result := (calc(_p, t) - _p[1]) / Cres + Vcres_0;
end;

function TDOpen.fCalcIh(t : precision) : precision;
begin
 Result := calc(par(a, b), t);
end;

function TDOpen.fCalcI2(t : precision) : precision;
begin
 Result := 0;
end;





function TDShort.TApproxFkt.Calc(t : Precision) : Precision;
begin
 Result := owner.calc(params, t);
end;


procedure TDShort.doUpdate(prev : TSRCValues);
var
  z     : Precision;
  Vmid  : Precision;
  Voutd : Precision; // reverse-rectified
  Cr    : Precision;
  d1, d2, d3 : Precision;
  action     : TSwitchAction;
  _p : TParamSet;
begin
 inherited;
 // case dependant values
 if (state.Q1 <> blocking) then begin Vmid := Vin; Cr := Cres; end
 else if (state.Q2 <> blocking) then begin Vmid := 0; Cr := Cres; end
 else begin Vmid := Vq2_0; Cr := 1/(1/Cres + 1/(Cq1+Cq2)); end;
 if (DiodesForward) then Voutd := Vout else Voutd := -Vout;
 // differential equation parameters
 _A := Cr*(Lh*Ls1 + Lh*Ls2 + Ls1*Ls2);
 _B := Cr*R2*(Lh + Ls1);
 _C := Lh + Ls2;
 _D := R2;
 p := (9*_A*_C - 3*_B**2)/(9*_A**2);
 q := (2*_B**3 - 9*_A*_B*_C + 27*_A**2*_D)/(27*_A**3);
 z := -(q/2) + sqrt((q/2)**2 + (p/3)**3); u := (abs(z))**(1/3)*sig(z);
 z := -(q/2) - sqrt((q/2)**2 + (p/3)**3); v := (abs(z))**(1/3)*sig(z);
 beta1 := u+v - _B/(3*_A);
 beta2 := -(1/2)*(u+v) - _B/(3*_A);
 omega := sqrt(3)/2 *(u-v);
 d := - Voutd/R2;
 b := - (Lh**2*beta1*(Ls1 + Ls2)*(beta1 - beta2)*(beta2**2 + omega**2)*(R2*Ih_0 - Voutd) + Lh*(Ls1*(Ls2*beta1*(beta1 - beta2)*(beta2**2 + omega**2)*(R2*(-I2_0 + Ih_0) - 2*Voutd) + R2*beta2*(beta1**2 - beta2**2 - omega**2)*(-R2*I2_0 - Voutd)) + Ls2**2*beta1*(beta1 - beta2)*(beta2**2 + omega**2)*(-R2*I2_0 - Voutd) + Ls2*R2*beta2*(beta1**2 - beta2**2 - omega**2)*(-R2*I2_0 - Voutd) + R2**2*(-R2*I2_0 - Vcres_0 + Vmid - Voutd)*(beta1*beta2 - beta2**2 - omega**2)) + Ls1*(Ls2*beta1 + R2)*(Ls2*(beta1 - beta2)*(beta2**2 + omega**2) + R2*(beta1*beta2 - beta2**2 - omega**2))*(-R2*I2_0 - Voutd))/(R2**2*omega*(beta1**2 - 2*beta1*beta2 + beta2**2 + omega**2)*(Lh*(Ls1 + Ls2) + Ls1*Ls2));
 c := (Ih_0*Lh - Voutd/R2*(R2/beta1 + Ls2 + Lh) + R2*b*omega/(beta2**2 + omega**2) - I2_0*(R2/beta1 + Ls2))/(R2*beta2/(beta2**2 + omega**2) - R2/beta1);
 a := -c - d + I2_0;
 a_ := a/Lh*(R2/beta1 + Ls2);
 b_ := (Ls2*b + (R2*b*beta2 + R2*c*omega)/(beta2**2 + omega**2))/Lh;
 c_ := (Ls2*c + (- R2*b*omega + R2*c*beta2)/(beta2**2 + omega**2))/Lh;
 d_ := -d;
 // duration
 action := GetNextSwitchAction(t_offset, endTime);
 d1 := getFirstZero(par(a, b, c), -d, not DiodesForward);
 if (state.Q1 = channel) then begin
  if (action = Q1_open) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := true; end;
  if (d1 < duration) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := false; duration := d1 end;
 end else if (state.Q2 = channel) then begin
  if (action = Q2_open) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := true; end;
  if (d1 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := channel; nextState.D := false; duration := d1 end;
 end else if (state.Q1 = diode) then begin
  d2 := getFirstZero(par(a+a_, b+b_, c+c_), 0, true);
  if (action = Q1_open) then begin nextState.Q1 := diode; nextState.Q2 := blocking; nextState.D := true; end;
  if (action = Q1_close) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := true; end;
  if (action = Q2_close) then begin nextState.Q2 := channel; nextState.Q1 := blocking; nextState.D := true; end;
  if (d1 < duration) then begin nextState.Q1 := diode; nextState.Q2 := blocking; nextState.D := false; duration := d1 end;
  if (d2 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := true; duration := d2 end;
 end else if (state.Q2 = diode) then begin
  d2 := getFirstZero(par(a+a_, b+b_, c+c_), 0, false);
  if (action = Q2_open) then begin nextState.Q1 := blocking; nextState.Q2 := diode; nextState.D := true; end;
  if (action = Q2_close) then begin nextState.Q1 := blocking; nextState.Q2 := channel; nextState.D := true; end;
  if (action = Q1_close) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := false; end;
  if (d1 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := diode; nextState.D := false; duration := d1 end;
  if (d2 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := true; duration := d2 end;
 end else begin
  _p := integrate(par(a+a_, b+b_, c+c_));
  d2 := getFirstZero(_p, (Vq2_0 - 0) * (Cq1+Cq2) + (_p[0]+_p[2]), true);
  d3 := getFirstZero(_p, (Vq2_0 - Vin) * (Cq1+Cq2) + (_p[0]+_p[2]), false);
  if (action = Q1_close) then begin nextState.Q1 := channel; nextState.Q2 := blocking; nextState.D := true; end;
  if (action = Q2_close) then begin nextState.Q1 := blocking; nextState.Q2 := channel; nextState.D := true; end;
  if (d1 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := blocking; nextState.D := false; duration := d1; end;
  if (d2 < duration) then begin nextState.Q1 := blocking; nextState.Q2 := diode; nextState.D := true; duration := d2; end;
  if (d3 < duration) then begin nextState.Q1 := diode; nextState.Q2 := blocking; nextState.D := true; duration := d3; end;
 end;
 durationMarker.Position := duration + t_offset;
 log('   next state: '+nextState.toString+'    d=' + floattostr(duration));
 // charge
 _p := integrate(par(a, b, c));
 I2_charge := I2_charge + Abs(calc(_p, duration) - _p[0] - _p[2] + d*duration);
 _p := integrate(par(a_, b_, c_));
 Ih_charge := Ih_charge + (calc(_p, duration) - _p[0] - _p[2] + d_*duration);
end;

function TDShort.getFirstZero(params : TParamSet; pd : Precision; gt : Boolean) : Precision;
var
 f, df : TApproxFkt;
 i : integer;
 s : Precision;
begin
 if (params[0] + params[2] > pd+Vin/10000) and (gt) or (params[0] + params[2] < pd-Vin/10000) and (not gt) then begin Result := 0; exit; end; // ist bereits zu Beginn erreicht
 f := TApproxFkt.Create;
 f.Owner := self;
 f.params := params;
 df := TApproxFkt.Create;
 df.Owner := self;
 df.params := derivate(params);
 i := 0;
 if (gt) then s := 1 else s := -1;
 while (i < 4) do begin
  Result := i * PI/(2*omega) + 1E-20;
  if (f.Newton(pd, df, Result, i * PI/(2*omega), (i+1) * PI/(2*omega), s)) then break;
  i := i+1;
 end;
 if (i = 4) then Result := Tinfinity; // wird nie erreicht
 f.Destroy;
 df.Destroy;
end;

function TDShort.derivate(const params : TParamSet) : TParamSet;
begin
 Result[0] := params[0]*beta1;
 Result[1] := params[1]*beta2 - params[2]*omega;
 Result[2] := params[1]*omega + params[2]*beta2;
end;

function TDShort.integrate(const params : TParamSet) : TParamSet;
begin
 Result[0] := params[0]/beta1;
 Result[1] := (params[1]*beta2 + params[2]*omega)/(beta2**2 + omega**2);
 Result[2] := (-params[1]*omega + params[2]*beta2)/(beta2**2 + omega**2);
end;

function TDShort.calc(const params : TParamSet; t : Precision) : Precision;
begin
 Result := params[0]*exp(beta1*t) + (params[1]*sin(omega*t) + params[2]*cos(omega*t))*exp(beta2*t);
end;

function TDShort.par(pa, pb, pc : Precision) : TParamSet;
begin
 Result[0] := pa;
 Result[1] := pb;
 Result[2] := pc;
end;

function TDShort.fCalcVq2(t : precision) : precision;
var _p : TParamSet;
begin
 if (state.Q1 <> blocking) then Result := Vin
 else if (state.Q2 <> blocking) then Result := 0
 else begin
  _p := integrate(par(a+a_, b+b_, c+c_));
  Result := Vq2_0 - (calc(_p, t) - (_p[0]+_p[2])) / (Cq1+Cq2);
 end;
end;

function TDShort.fCalcVLh(t : precision) : precision;
begin
 Result := Lh * calc(derivate(par(a_, b_, c_)), t);
end;

function TDShort.fCalcVcres (t : Precision) : Precision;
var _p : TParamSet;
begin
 _p := integrate(par(a+a_, b+b_, c+c_));
 Result := (calc(_p, t) - (_p[0]+_p[2])) / Cres + Vcres_0;
end;

function TDShort.fCalcIh(t : precision) : precision;
begin
 Result := calc(par(a_, b_, c_), t) + d_;
end;

function TDShort.fCalcI2(t : precision) : precision;
begin
 Result := calc(par(a, b, c), t) + d;
end;

end.

